# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE SETUP.PY
short_version = '4.19.0'
version = '4.19.0'
full_version = '4.19.0'
git_revision = '551dfb11841c0be176be4b6ebe7678aaa3bad68a'
release = True

if not release:
    version = full_version
    short_version += ".dev"
